#!/usr/bin/perl -w
# txt2graph (c) Andreas Romeyke,
# distributed under terms of GNU General Public License
# for license and details, check included POD and file COPYING

=pod

=head1 NAME

txt2graph - generates a dot-file from text-dependencies

=head1 DOWNLOAD

try L<http://andreas-romeyke.de/privat/txt2graph/>,
the changelog is on L<http://andreas-romeyke.de/privat/txt2graph/CHANGES>,

=cut

########################## START VERSION ##################################

use constant VERSION=>"1.2.3";

=pod

latest tarball is on L<http://andreas-romeyke.de/privat/txt2graph/txt2graph1.2.2.tgz>

=cut

############################ END VERSION ##################################

=pod

=head1 SYNOPSIS

B<txt2graph>  visualized  the structure and dependencies of a text readed
from stdin and generates a dot-File for B<graphviz> to stdout.

=head1 DEPENDENCIES

=over 4

=item *

You need already installed GraphViz tools, check L<http://www.graphviz.org> or L<http://www.research.att.com/sw/tools/graphviz/>

=item *

You need the perl-module L<GraphViz.pm> by Leon Brocard (since version 1.2). You can get it from CPAN.org, simply by entering CPAN-shell via C<'perl -MCPAN -e shell'> and install it by typing C<'install GraphViz'>, the L<GraphViz> needs additional modules, too.

=back

=head1 DESCRIPTION

=head2 What it does

B<txt2graph> should visualisize the structure and dependencies of a text,
because I needed it to analyze and understand the principles how my own
text-compressor should work for good results. With B<txt2graph> it is
possible to find the weights in a document. This tool is also useful to
compare documents in a similar way a B<ngram>-tool will do. With usage of
filterfile- and frequency-option you get little "topic"-maps of documents.

If you find another usage, please contact me.

=head2 How it works

B<txt2graph>  reads a text document from stdin, removes all non-alphas and
generates  an  array  (list) of words. Then it converts german-umlauts,
because  B<graphviz>  can only handle clean ASCII as node-description and
output a dot-file for a directed or an undirected graph.

=head2 Warning

The generated dot-file will be typically 4-5fold of original document.
This means also that L<dot> or L<neato> consumes very much time.

=head1 AUTHOR and COPYRIGHT

B<txt2graph> (c)2001-2003 by Andreas Romeyke (andreas.romeyke@web.de)

=head1 LICENSE

B<txt2graph>  is  distributed  under  the terms of the GNU General Public
License, you find a copy in this distribution in file COPYING or at L<http://www.gnu.org>.

Please check twice if you want a B<special license for commercial use> and contact me
with signed email (signed with PGP or GnuPG!) and the subject "txt2graph license".

=head1 THANKS

A  special greeting goes to my friend-girl Maren for her patience, to
Matthias Richter for his ideas, Leon Brocard for his GraphViz-Module
and to Derek Jones for his bugreports. Last but not least a greeting
goes to the team around the GraphViz-tools.

=head1 HOW TO USE IT

=over 4

=item * for help simple type:

C<perl txt2graph.pl --help>

=item * for manual simple type:

C<perldoc txt2graph.pl>

=item * to   create  a  weighted  undirected  graph:

C<cat  mytext.txt|perl txt2graph.pl  --undirect  E<gt>  mytext.dot>

C<neato -Tps mytext.dot E<gt> mytext.ps>

Or quite simpler (since version 1.2):

C<cat mytext.txt| perl txt2graph.pl --undirect --format=ps E<gt> mytest.ps>

=item * to  create  an  undirected graph:

C<cat mytext.txt|perl txt2graph.pl --undirect E<gt> mytext.dot>

C<dot -Tps mytext.dot E<gt> mytext.ps>

or quite simpler (since version 1.2):

C<cat mytext.txt| perl txt2graph.pl --undirect --format=png E<gt> mytext.png>

=item * to  create  a  directed  graph:

C<cat  mytext.txt|perl txt2graph.pl --direct E<gt> mytext.dot>

C<dot -Tps mytext.dot E<gt> mytext.ps>

=item * create a directed graph with twopi-layout as SVG:

C<cat mytext.txt| perl txt2graph.pl --direct --format=svg --layout=twopi E<gt> mytext.svg>

(since version 1.2)

=item * you can avoid words by using stoplist, minfrequency, maxfrequency, mincount and maxcount, more information via

C<perl --help | more>

=item * a more complex example is below:

C< txt2graph.pl --infile=TODO --concentrate --layout=dot --format=ps  --zeroword --records --casesensitive --minwordlen=3 --remove-multiples --outfile=test.ps>

It reads the file 'TODO', builds an PostScript-file 'test.ps'. The nodes were clustered, that means strong related words were nearby. The graph is undirected, multiple links were supressed and filtered words were substituted by a special node. The words must be greater than 3 characters an will be used in casesensitive way

=back

=head1 BUGS

There were following known bugs:

=over 4

=item *

older version of GraphViz coredump if nodescount is high. This is not a failure in this script, but	has an effect on it. Use C<'--format=canon'> to recheck if this script fails too.

=back

Of course, there are some additional bugs, but not known yet. If you detect one,
please do not hesitate to contact me at andreas.romeyke@web.de

=head1 TODO

check extra file L<http://andreas-romeyke.de/privat/txt2graph/TODO>. If you have suggestions, or patches (prefered), contact me, too.

=head1 API Description

... coming soon ...

=cut

use strict;
use Getopt::Long;
use GraphViz;
#use Data::Dumper;
$Getopt::Long::autoabbrev=1;
$Getopt::Long::order=$REQUIRE_ORDER;


#
# Default values for options
#
my %opt=();
${opt{'infile'}}="";
${opt{'outfile'}}="";
${opt{'format'}}="canon";
${opt{'directgraph'}}= -1;
${opt{'undirectgraph'}}= -1;
${opt{'filterfile'}}="";
${opt{'sortalpha'}}=1;
${opt{'sortfreq'}}=0;
${opt{'removemultiples'}}=0;
${opt{'concentrate'}}=0;
${opt{'fontsize'}}=10;
${opt{'casesensitive'}}=1;
${opt{'soundex'}}=0;
${opt{'maxfrequency'}}=1;
${opt{'minfrequency'}}=0;
${opt{'maxcount'}}=0;
${opt{'mincount'}}=0;
${opt{'maxwordlen'}}=0;
${opt{'minwordlen'}}=0;
${opt{'debug'}}=0;
${opt{'zeroword'}}=1; #a placeholder for filtered words
${opt{'freqfile'}}="";
${opt{'layout'}}='dot';
${opt{'records'}}=0;
#
# Used Variables
#

# Filehandles
#INFILE;
#OUTFILE;
#FFILE;

# Strings
my $gr_direction;
my $gr_graph;

# Buffer
my @buffer_words=();
my %buffer_filter=();
my @buffer_set=();
my %buffer_freq=();
my $all_freq=0; #needed for freqency calculation
my %buffer_nodes=();
my %buffer_sort=();
# tmp Vars
my $last="";
my $last2="";
my $num=0;
my $nodecounter=0;


# print Copyleftinfos
sub printcopyleft() {
	print STDERR "txt2graph v",VERSION," (c)2001-2003 by Andreas Romeyke\n";
	print STDERR "distributed under the terms of Gnu General Public License\n";
	print STDERR "check the copy in distribution or check http://www.gnu.org\n";
	print STDERR "\n";
	print STDERR "please contact me: andreas{dot}romeyke[at]web{dot}de\n";
	print STDERR "or visit: http://andreas-romeyke.de/privat\n\n";
}


#
# DEBUG
#
sub debugswitch() {
	if (${opt{'debug'}}==1) {
		print STDERR "Switches are:\n";
		print STDERR "\tlayout:\t\t\t${opt{'layout'}}\n";
		print STDERR "\tformat:\t\t\t${opt{'format'}}\n";
		print STDERR "\tinfile:\t\t\t${opt{'infile'}}\n";
		print STDERR "\toutfile:\t\t${opt{'outfile'}}\n";
		print STDERR "\tdirectgraph:\t\t${opt{'directgraph'}}\n";
		print STDERR "\tundirectgraph:\t\t${opt{'undirectgraph'}}\n";
		print STDERR "\tfilterfile:\t\t${opt{'filterfile'}}\n";
		print STDERR "\tsortalpha:\t\t\t${opt{'sortalpha'}}\n";
		print STDERR "\tsortfreq:\t\t\t${opt{'sortfreq'}}\n";
		print STDERR "\tremove multiples:\t${opt{'removemultiples'}}\n";
		print STDERR "\tconcentrate:\t${opt{'concentrate'}}\n";
		print STDERR "\tfontsize:\t\t${opt{'fontsize'}}\n";
		print STDERR "\tcasesensitive:\t\t${opt{'casesensitive'}}\n";
		print STDERR "\tsoundex:\t\t${opt{'soundex'}}\n";
		print STDERR "\tmaxfrequency:\t\t${opt{'maxfrequency'}}\n";
		print STDERR "\tminfrequency:\t\t${opt{'minfrequency'}}\n";
		print STDERR "\tmaxcount:\t\t${opt{'maxcount'}}\n";
		print STDERR "\tmincount:\t\t${opt{'mincount'}}\n";
		print STDERR "\tmaxwordlen:\t\t${opt{'maxwordlen'}}\n";
		print STDERR "\tminwordlen:\t\t${opt{'minwordlen'}}\n";
		print STDERR "\tzeroword:\t\t${opt{'zeroword'}}\n";
		print STDERR "\trecords:\t\t${opt{'records'}}\n";
		print STDERR "\tdebug:\t\t\t${opt{'debug'}}\n";
	}
}


# make logical tests with options
sub optionlogic() {
	if ($Getopt::Long::error== 1) { #Errors von Getopt selber.
		die "Error in Options\n";
	}
	if (
		(${opt{'layout'}} ne "dot")&&
		(${opt{'layout'}} ne "neato")&&
		(${opt{'layout'}} ne "twopi")
		) {
		die "Layout must be 'dot', 'neato' or 'twopi'\n";
	}
	if (
		(${opt{'format'}} ne "canon")&&
		(${opt{'format'}} ne "png")&&
		(${opt{'format'}} ne "ps")&&
		(${opt{'format'}} ne "txt")&&
		(${opt{'format'}} ne "gd")&&
		(${opt{'format'}} ne "gif")&&
		(${opt{'format'}} ne "jpeg")&&
		(${opt{'format'}} ne "fig")&&
		(${opt{'format'}} ne "svg")
		) {
			die "output format must be one of 'canon', 'png', 'ps', 'txt', 'gd', 'gif', 'jpeg', 'fig', 'svg', not ${opt{'format'}}\n";
		}
	if (${opt{'outfile'}} ne "") {
		if (${opt{'infile'}} eq ${opt{'outfile'}}) {
			die "Infile \"${opt{'infile'}}\" and Outfile \"${opt{'outfile'}}\" are same\n";
		}
		if (${opt{'filterfile'}} eq ${opt{'outfile'}}) {
			die "Filterfile \"${opt{'filterfile'}}\" and Outfile \"${opt{'outfile'}}\" are same\n";
		}
		if (${opt{'freqfile'}} eq ${opt{'outfile'}}) {
			die "Freqfile \"${opt{'freqfile'}}\" and Outfile \"${opt{'outfile'}}\" are same\n";
		}
	}
	if (${opt{'freqfile'}} ne "") {
		if (${opt{'infile'}} eq ${opt{'freqfile'}}) {
			die "Infile \"${opt{'infile'}}\" and Freqfile \"${opt{'freqfile'}}\" are same\n";
		}
		if (${opt{'filterfile'}} eq ${opt{'freqfile'}}) {
			die "Filterfile \"${opt{'filterfile'}}\" and Freqfile \"${opt{'freqfile'}}\" are same\n";
		}
		if (${opt{'freqfile'}} eq ${opt{'outfile'}}) {
			die "Freqfile \"${opt{'freqfile'}}\" and Outfile \"${opt{'outfile'}}\" are same\n";
		}
	}
	#
	# direction
	#
	if (
			(${opt{'undirectgraph'}}!= -1)&&
			(${opt{'directgraph'}}!= -1)&&
			((${opt{'directgraph'}} ^ ${opt{'undirectgraph'}}) == 0)
		) {
		die "directgraph and undirectgraph is set\n";
	}
	if (
		(${opt{'directgraph'}}==1)&&
		(${opt{'undirectgraph'}}<=0)
		) {
		$gr_direction="forward";
	} else {
		$gr_direction="none";
	}

	if (
		(${opt{'maxcount'}} > 0) &&
		(${opt{'maxcount'}} < ${opt{'mincount'}})
		) {
		die "mincount is greater than maxcount\n";
	}
	if (${opt{'mincount'}} < 0) {
		die "mincount ${opt{'mincount'}} is less than 0\n";
	}
	if (
		(${opt{'maxcount'}} > 0)&&
		(${opt{'mincount'}} == 0)
		) {
		die "mincount must be set, if maxcount is used\n";
	}
	if (
		(${opt{'maxfrequency'}} > 0)&&
		(${opt{'maxfrequency'}} < ${opt{'minfrequency'}})
		) {
		die "minfrequency is greater than maxfrequency\n";
	}
	if (${opt{'minfrequency'}} < 0) {
		die "minfrequency ${opt{'minfrequency'}} is less than 0\n";
	}
	if (${opt{'maxfrequency'}} > 1) {
		die "maxfrequency must be between 0 and 1\n";
	}
	if (
		(${opt{'maxwordlen'}} > 0)&&
		(${opt{'maxwordlen'}} <= ${opt{'minwordlen'}})
		) {
		die "minwordlen is greater than maxwordlen\n";
	}
	if (
		(${opt{'maxwordlen'}} > 0)&&
		(${opt{'minwordlen'}}==0)
		) {
		die "minwordlen must be set, if maxwordlen is used\n";
	}
	if (${opt{'minwordlen'}} < 0) {
		die "minwordlen ${opt{'minwordlen'}} is less than 0\n";
	}
}


sub parseoptions(){
	&GetOptions(
			"layout=s" => \${opt{'layout'}},
			"format=s" => \${opt{'format'}},
			"infile=s" => \${opt{'infile'}},
			"outfile=s" => \${opt{'outfile'}},
			"print-manual" => sub {manual();},
			"help" => sub {help();},
			"direct-graph!" => \${opt{'directgraph'}},
			"undirect-graph!" => \${opt{'undirectgraph'}},
			"filterfile=s" => \${opt{'filterfile'}},
			"freqfile=s" => \${opt{'freqfile'}},
			"sortalpha" => sub {${opt{'sortalpha'}}=1; ${opt{'sortfreq'}}=0;},
			"sortfreq" => sub {${opt{'sortalpha'}}=0; ${opt{'sortfreq'}}=1;},
			"remove-multiples!" => \${opt{'removemultiples'}},
			"concentrate" => \${opt{'concentrate'}},
			"fontsize=i" => \${opt{'fontsize'}},
			"casesensitive!" => \${opt{'casesensitive'}},
			"soundex!" => \${opt{'soundex'}},
			"maxfrequency=f" => \${opt{'maxfrequency'}},
			"minfrequency=f" => \${opt{'minfrequency'}},
			"maxcount=i" => \${opt{'maxcount'}},
			"mincount=i" => \${opt{'mincount'}},
			"minwordlen=i" => \${opt{'minwordlen'}},
			"maxwordlen=i" => \${opt{'maxwordlen'}},
			"zeroword!" => \${opt{'zeroword'}},
			"debug!" => \${opt{'debug'}},
			"records" => \${opt{'records'}}
	);
	debugswitch();
	optionlogic();
}


#
# Wenn Hilfe
# if help
#
sub help() {
	print "txt2graph visualized the structure and dependencies of a text readed from stdin and generates a dot-File for graphviz to stdout.\n\n";
	print "Switches are:\n";

	print "\t--casesensitive\t\twords are handled casesensitive\n\t\t\t\t(D:true)\n";
	print "\t--debug\t\t\tprints debugging information\n\n";
	print "\t--directgraph\t\tprints directed graph\n";
	print "\t--concentrate\t\tclusters strong dependend words\t\t\t\t(D:false)\n";
	print "\t--filterfile=<FILE>\twords (per line) from this file will be filtered\n";
	print "\t--fontsize\t\tsets the size of node-labels\n\t\t\t\t(D: 10 pt)\n";
	print "\t--format=<OUTPUTFORMAT>\t\twhere <OUTPUTFORMAT> could be  one of 'canon', 'png', 'ps', 'txt', 'gd', 'gif', 'jpeg', 'fig', 'svg'\n\t\t\t\t(D: canon)\n";
	print "\t--help\t\t\tprints this help\n";
	print "\t--infile=<FILE>\t\tread words from file\n\t\t\t\t(D: stdin)\n";
	print "\t--layout=<LAYOUT>\t\twhere <LAYOUT> could be 'dot', 'neato', or 'twopi'\n\t\t\t\t(D: dot)\n";
	print "\t--maxcount=<VALUE>\tit filters dependencies with a \n\t\t\t\tcount higher than value\n";
	print "\t--maxfrequency=<VALUE>\tit filters dependencies with a \n\t\t\t\tfrequency higher than value\n\t\t\t\t(D: 1)\n";
	print "\t--maxwordlen=<VALUE>\tit filters words with a length more than value\n";
	print "\t--mincount=<VALUE>\tit filters dependencies with a \n\t\t\t\tcount lower than value\n";
	print "\t--minfrequency=<VALUE>\tit filters dependencies with a \n\t\t\t\tfrequency lower than value\n\t\t\t\t(D: 0)\n";
	print "\t--minwordlen=<VALUE>\tit filters words with a length less than value\n";
	print "\t--nocasesensitive\topposite of --casesensitive\n\t\t\t\t(D: false)\n";
	print "\t--nodirectgraph\t\tprints undirected graph\n";
	print "\t--nosoundex\t\topposite of --soundex\n\t\t\t\t(D: true)\n";
	print "\t--nozeroword\t\topposite of --zeroword\n\t\t\t\t(D: false)\n";
	print "\t--outfile=<FILE>\twrite dot/text/picture-file\n\t\t\t\t(D: stdout)\n";
	print "\t--print-manual\t\tprints builtin manual with \"perldoc\"\n";
	print "\t--records\t\t\tdisplays additional information per node,\n\t\t\t\tusefull if you want --zerowords\n\t\t\t\t(D:false)\n";
	print "\t--remove-multiples\tmultiple dependencies are counted as one\n\t\t\t\t(D: false)\n";
	print "\t--sortalpha\t\tsort dependencies alphabetically\n\t\t\t\t(excludes --sortfreq)\n\t\t\t\t(D: true)\n";
	print "\t--sortfreq\t\tsort dependencies after frequency\n\t\t\t\tcould help to speedup neato (excludes --sortalpha)\n\t\t\t\t\n\t\t\t\t(D: false)\n";
	print "\t--soundex\t\tconvert words in \"spoken\" equivalents\n\t\t\t\t(D: false)\n";
	print "\t--undirectgraph\t\tprints undirected graph\n";
	print "\t--zeroword\t\tfiltered words represented by a central placeholder,\n\t\t\t\tif unset dependency graph will be shown as isolated clusters\n\t\t\t\t(D: true)\n";
	print "Today the switches \"--soundex\" and \"--concentrate\" are unsupported.\n";
	print "You can use abbreviations, this tool will help you if they are inconsistent.\n";
	print "Words of a document are parsed in following chain:\n";
	print "\t-> loading document\n";
	print "\t-> elimination of non-alpha\n";
	print "\t-> casesensitivity\n";
	print "\t-> wordlenfilter\n";
	print "\t-> translation of special german chars\n";
	print "\t-> frequency-counter for words\n";
	print "\t-> wordbased-filter\n";
	print "\t\t-> counted words\n";
	print "\t\t-> frequency of words\n";
	print "\t\t-> elimination of all words in filterfile\n";
	print "\t\t-> soundex-converter\n";
	print "\t-> linkbuilding\n";
	print "\t-> frequency-counter for links\n";
	print "\t-> linkbased-filter\n";
	print "\t\t-> sort by alpha\n";
	print "\t\t-> sort by frequency, most frequented first\n";
	print "\t\t-> remove-multiples\n";
	print "\t-> graph-generator\n\n";

	print "For a very detailed description, how you can use it, display\n";
	print "the builtin manual with \"perldoc txt2graph\", create as HTML\n";
	print "with \"pod2html txt2graph\", as manual \"pod2man txt2graph\",\n";
	print "or extract it as LaTeX-Document with \"pod2latex txt2graph\"\n";
	print "(to PS: \n";
	print " \"latex txt2graph.tex\",\n";
	print " \"cat txt2graph.dvi|dvips -f -D600 -mode docutech >txt2graph.ps\")\n\n";
	print "Graphviz is available \@ http://www.research.att.com/sw/tools/graphviz/\n\n";
	print "example:\n";
	print "\tcat mytext.txt|txt2graph --direct > mytext.dot\n";
	print "\tdot -Tps mytext.dot > mytext.ps\n\n";
	print "please contact me: andreas{dot}romeyke[at]web{dot}de\n";
	print "or visit: http://andreas-romeyke.de/privat\n";
	exit 0;
}


#
# print POD documentation
#
sub manual() {
	system('perldoc', $0);
	exit 0;
}


sub fileprep() {
	#
	# possible INFILE redirection
	#
	if (${opt{'infile'}} eq "") { #setze STDIN
		open INFILE, "-" or die "could not reopen STDIN\n";
	} else {
		open INFILE, "<${opt{'infile'}}" or die "infile \"${opt{'infile'}}\" could not be opened\n";
	}
	#
	# possible OUTFILE redirection
	#
	if (${opt{'outfile'}} eq "") { #setze STDOUT
		open OUTFILE, ">-" or die "could not reopen STDOUT\n";
	} else {
		open OUTFILE, ">${opt{'outfile'}}" or die "outfile \"${opt{'outfile'}}\" could not be opened\n";
	}
	#
	# possible FFILE usage
	#
	if (${opt{'filterfile'}} ne "" ) {
		open FFILE, "<${opt{'filterfile'}}" or die "filterfile \"${opt{'filterfile'}}\" could not be opened\n";
		#fill filterbuffer
		while (<FFILE>) {
			s/\n//;
			$buffer_filter{$_}=1;
		}
	}
}


# read words and sorts them by special rules, pE. wordlen{min,max}
# in future versions rules could be regEx too
#
# Wrter lesen und nach Filterregeln (wordlen{min,max}, aussortieren
# spter knnten spezielle RegEx-Regeln mglich sein.
# @buffer_words
#
sub readin() {
	push @buffer_words, "B E G I N";
	while (<INFILE>) {
		#readline and eliminate all newlines and special chars
		#Zeile einlesen, dabei alle newlines, und Sonderzeichen eliminieren
	#    s/  +/ /g;
		study; # beschleunigt Abarbeitung der ff. Ersetzungen / improved performance
		tr/[^A-Za-z]+/ /cs;# alle nicht Woerter loeschen / delete all non-word chars
		if (${opt{'casesensitive'}}==0) {
			tr/A-Z/a-z/;   # Gross- zu Kleinschreibung wandeln / upper to lower case
		}
		if ((${opt{'maxwordlen'}} >0)&&(${opt{'minwordlen'}} >0)) {
			# Filter fuer wordlen / wordlen filter
			s/\b[A-Za-z]{1,${opt{'minwordlen'}}}\b//g;
			s/\b[A-Za-z]{${opt{'maxwordlen'}},}\b//g;
		}

#		s//Ae/g; # umlaut substitution because older graphviz could not handle these
#		s//ae/g;
#		s//Oe/g;
#		s//oe/g;
#		s//Ue/g;
#		s//ue/g;
#		s//sz/g;

		#    print "... $_\n";
		push @buffer_words, split (/\s+/);
	}
	push @buffer_words, "E N D";
	push @buffer_words, "E N D";
}


sub writefreqtable() {
	## for debugging
	open (FREQ, ">.freq") or die;
	foreach $_ (keys %buffer_freq) {
		print FREQ $_," -> ",$buffer_freq{$_}/$all_freq,"\n";
	}
	close FREQ;
}


#
# Hufigkeitsbestimmung der Wrter
# count frequencies of words
sub countwords() {
		foreach $_ (@buffer_words) {
			if (not defined $buffer_freq{$_}) {
				$buffer_freq{$_}=1;
			} else {
				$buffer_freq{$_}++;
			}
			$all_freq++;
		}
	writefreqtable();
}


#
# sort out by counts
# nach Anzahl filtern
#
sub cleanbycounts() {
	if (${opt{'maxcount'}}>=1) {
		my @buffer_tmp=();
		foreach $_ (@buffer_words) {
			if (
				($buffer_freq{$_} >= ${opt{'mincount'}}) &&
				($buffer_freq{$_} <= ${opt{'maxcount'}})
			) {
				push @buffer_tmp, $_;
			} else {
				if (${opt{'zeroword'}}==1) {
					push @buffer_tmp, 'U N K N O W N';
				}
			}
		}
		@buffer_words= @buffer_tmp;
	}
}


#
# sort out by frequencies
# nach Hufigkeiten filtern
#
sub cleanbyfreqs() {
	if (${opt{'maxfrequency'}}>0) {
		my @buffer_tmp=();
		foreach $_ (@buffer_words) {
			if (
				(@buffer_words > 0)&&
				(($buffer_freq{$_}/@buffer_words) >= ${opt{'minfrequency'}}) &&
				(($buffer_freq{$_}/@buffer_words) <= ${opt{'maxfrequency'}})
			) {
				push @buffer_tmp, $_;
			} else {
				if (${opt{'zeroword'}}==1) {
					push @buffer_tmp, 'U N K N O W N';
				}
			}
		}
		@buffer_words= @buffer_tmp;
	}
}


#
# eliminate all words from @buffer_words which exists in %buffer_filter
# Eliminieren aller Wrter aus @buffer_words, die in %buffer_filter vorkommen
#
sub cleanbystoplist() {
	if (${opt{'filterfile'}} ne "") {
		my @buffer_tmp=();
		foreach my $tmp (@buffer_words) {
			if ((exists $buffer_filter{$tmp}) && ($buffer_filter{$tmp}==1)) {
				if (${opt{'zeroword'}}==1) {
					push @buffer_tmp, 'U N K N O W N';
				}
			} else {
				push @buffer_tmp, $tmp;
			}
		}
		@buffer_words= @buffer_tmp;
	}
}

# TODO: Soundex - Konvertierung / soundex - convertion
#
sub convertbysoundex() {
}


#
# semantic analysis (not really, means here: cluster 2 neighbored words)
# Satzanalyse (hier: Zusammenfassen je zweier aufeinanderfolgende Worte)
#
sub buildlinks() {
	$last="";
	$last2="";
	$num=0;
	$nodecounter=0;
	foreach $_ (@buffer_words) {
		if (! exists $buffer_nodes{$_}) {
			$buffer_nodes{$_}=$nodecounter; # Nodeshash aufbauen / create nodehash
			$buffer_sort{$nodecounter}=$_; #Hash fuer Sortierung aufbauen / create hash to sort things
			$nodecounter++;
		}
		$num=$num ^ 1;
		if ($num==1) {
			if (
				($_ ne "")&&
				($last ne "")&&
				($last2 ne "")
				) {
				push @buffer_set, "$buffer_nodes{$last2}\:\:$buffer_nodes{$last}";
				push @buffer_set, "$buffer_nodes{$last}\:\:$buffer_nodes{$_}";
			}
		}
		$last2=$last;
		$last=$_;
	}
}


#
# Hufigkeitsbestimmung der Wortvektoren/links
# count frequencies of word combinations/links
#
sub countlinks() {
	if (
		(${opt{'removemultiples'}}!=0)||
		(${opt{'concentrate'}}!=0)
		) {
		# delete previous used all_freq and %buffer_freq
		$all_freq=0; %buffer_freq=();
		foreach $_ (@buffer_set) {
			$buffer_freq{$_}++;
			$all_freq++;
		}
	}
}


#
# sort if wanted
# ev. Sortieren
#
sub sortbyalpha() {
	if (${opt{'sortalpha'}}==1) {
		my @buffer_tmp=();
	#	@buffer_tmp=sort mysort @buffer_set;
		@buffer_tmp=sort {
			my ($a1, $a2)=split /::/, $a;
			my ($b1, $b2)=split /::/, $b;
			$a1=$buffer_sort{$a1}."::".$buffer_sort{$a2};
			$b1=$buffer_sort{$b1}."::.".$buffer_sort{$b2};
			return ($a1 cmp $b1);
		} @buffer_set;
		@buffer_set=@buffer_tmp;
	}
}


sub sortbyfreq() {
};


#
# delete duplicate links
# Wiederholungen loeschen
sub cleanduplicatelinks() {
	if (${opt{'removemultiples'}}!=0) {
		my %buffer_tmp=();
		foreach $_ (@buffer_set) {
			$buffer_tmp{$_}=1;
		}
		@buffer_set= keys %buffer_tmp;
	}
}


#
# output to dot
# Ausgabe zum eigentlichen Zeichnen
#
sub writedot() {

	my $dot=GraphViz->new(
	layout => ${opt{'layout'}},
		no_overlap=>1,
		node=>{
			fontsize=>${opt{'fontsize'}},
			concentrate=>${opt{'concentrate'}},
			shape=>"record",
		}
	);
	foreach $_ (keys %buffer_nodes) {
		my $myorig= " ";
		my $mycount=$buffer_freq{$_};
		$mycount=defined $mycount?$mycount:" ";
		if (
			($_ eq "B E G I N")||
			($_ eq "E N D")||
			($_ eq "U N K N O W N")
		) {
			if ($opt{'records'} == 0) {
				$dot->add_node($buffer_nodes{$_}, label=>"$_\n", style=>'bold');
			} else {
				$dot->add_node($buffer_nodes{$_}, label=>["$_\n","$mycount\n"], style=>'bold');
			}
		} else {
			if ($opt{'records'} == 0) {
				$dot->add_node($buffer_nodes{$_}, label=>"$_\n");
			} else {
				$dot->add_node($buffer_nodes{$_}, label=>["$_\n","$mycount\n"]);
			}
		}
	}
	# body
	foreach $_ (@buffer_set) {
		my $freq=$buffer_freq{$_};
		if (not defined $freq) {
			$freq=1;
		}
		my ($left,$right)=split /::/;
#print STDERR "$left,$right ---> $freq\n";
		$dot->add_edge($left => $right, dir=>$gr_direction, weight=>$freq );
	}
	my $format;
	if (${opt{'format'}} eq 'canon') {$format=$dot->as_canon;}
	if (${opt{'format'}} eq 'png') {$format=$dot->as_png;}
	if (${opt{'format'}} eq 'gif') {$format=$dot->as_gif;}
	if (${opt{'format'}} eq 'jpeg') {$format=$dot->as_jpeg;}
	if (${opt{'format'}} eq 'ps') {$format=$dot->as_ps;}
	if (${opt{'format'}} eq 'svg') {$format=$dot->as_svg;}
	if (${opt{'format'}} eq 'txt') {$format=$dot->as_text;}
	if (${opt{'format'}} eq 'gd') {$format=$dot->as_gd;}
	if (${opt{'format'}} eq 'fig') {$format=$dot->as_fig;}
	print OUTFILE $format;
}



sub wordbasedfilter() {
	cleanbycounts();
	cleanbyfreqs();
	cleanbystoplist();
	convertbysoundex();
}


sub linkbasedfilter() {
	sortbyalpha();
	sortbyfreq();
	cleanduplicatelinks();
}


###
### main program
###

printcopyleft();
parseoptions();
fileprep();
readin();
countwords();
wordbasedfilter();
buildlinks();
countlinks();
linkbasedfilter();
writedot();

