#!/usr/bin/perl -w
# wordtable (c) Andreas Romeyke,
# distributed under terms of GNU General Public License
# for license and details, check included POD and file COPYING

=pod

=head1 NAME

wordtable - reads from stdin and builds hash of words (with frequencies)

=cut
########################## START VERSION ##################################

use constant VERSION=>"0.0.1";

=pod

latest tarball is on L<http://andreas-romeyke.de/>

=cut

############################ END VERSION ##################################



=head1 HOW TO USE IT

=over 4

=item * for help simple type:

C<perl wordtable.pl --help>

=item * for manual simple type:

C<perldoc wordtable.pl>

=item * to   get longest word

C<cat  mytext.txt|perl wordtablepl E<gt> longest_words.txt>

=item * to  get most frequented words

C<cat mytext.txt|perl wordtable.pl E<gt> mytext.dot>

=cut

no utf8;
use strict;
use Getopt::Long;
$Getopt::Long::autoabbrev=1;
$Getopt::Long::order=$REQUIRE_ORDER;


#
# Default values for options
#
my %opt=();
${opt{'wordlen'}}=0;
${opt{'wordfreq'}}=1;
${opt{'minwordlen'}}=0;
${opt{'maxwordlen'}}=0;
${opt{'casesensitive'}}=0;
${opt{'infile'}}="";
${opt{'outfile'}}="";
${opt{'with-progress'}}=0;

#
# Used Variables
#
my %words=();

# print Copyleftinfos
sub printcopyleft() {
	print STDERR "wordtable v",VERSION," (c)2009 by Andreas Romeyke\n";
	print STDERR "distributed under the terms of Gnu General Public License\n";
	print STDERR "check the copy in distribution or check http://www.gnu.org\n";
	print STDERR "\n";
	print STDERR "please contact me: andreas{dot}romeyke[at]web{dot}de\n";
	print STDERR "or visit: http://andreas-romeyke.de/\n\n";
}

sub parseoptions(){
	&GetOptions(
		"print-manual" => sub {manual();},
		"help" => sub {help();},
		"wordlen" => sub {${opt{'wordlen'}}=1; ${opt{'wordfreq'}}=0;},
		"wordfreq" => sub {${opt{'wordlen'}}=0; ${opt{'wordfreq'}}=1;},
		"casesensitive" => sub {${opt{'casesensitive'}}=0;},
		"minwordlen" => sub {${opt{'minwordlen'}}=0;},
		"maxwordlen" => sub {${opt{'maxwordlen'}}=0;},
		"infile=s" => \${opt{'infile'}},
                "outfile=s" => \${opt{'outfile'}},
		"with-progress" => sub {${opt{'with-progress'}}=1-${opt{'with-progress'}};},
	);
}


		#
# Wenn Hilfe
# if help
		#
		sub help() {
			print "wordtable -- counts words in given file and sorts them by frequency or by wordlength\n\n";
			print "Switches are:\n";

			print "\t--wordlen (default: false)\n";
			print "\t--wordfreq (default: true)\n";
			print "\t--with-progress (default: false)\n";
			print "\t[--infile file] (default: stdin)\n";
			print "\t[--outfile file] (default: stdout)\n";
			print "\t--help\n";
			print "\t--print-manual\n";
			exit 0;
		}

		#
# print POD documentation
		#
		sub manual() {
			system('perldoc', $0);
			exit 0;
		}


		sub fileprep() {
			#
			# possible INFILE redirection
			#
			if (${opt{'infile'}} eq "") { #setze STDIN
			open INFILE, "-" or die "could not reopen STDIN\n";
		} else {
			open INFILE, "<${opt{'infile'}}" or die "infile \"${opt{'infile'}}\" could not be opened\n";
		}
		#
		# possible OUTFILE redirection
		#
		if (${opt{'outfile'}} eq "") { #setze STDOUT
		open OUTFILE, ">-" or die "could not reopen STDOUT\n";
	} else {
		open OUTFILE, ">${opt{'outfile'}}" or die "outfile \"${opt{'outfile'}}\" could not be opened\n";
	}
}


# read words and sorts them by special rules, pE. wordlen{min,max}
# in future versions rules could be regEx too
#
# Wrter lesen und nach Filterregeln (wordlen{min,max}, aussortieren
# spter knnten spezielle RegEx-Regeln mglich sein.
# @buffer_words
#


###
### main program
###

printcopyleft();
parseoptions();
fileprep();
$|=1; select(STDERR);
my %stats;
$stats{'lines'}=0;
$stats{'words'}=0;
$stats{'bytes'}=0;
$stats{'whitespaces'}=0;
$stats{'wordchars'}=0;
$stats{'starttime'}=time;
$stats{'uniqs'}=0;
while (<INFILE>) {
	#readline and eliminate all newlines and special chars
	#Zeile einlesen, dabei alle newlines, und Sonderzeichen eliminieren
	#    s/  +/ /g;
	study; # beschleunigt Abarbeitung der ff. Ersetzungen / improved performance
	$stats{'lines'}++;
	$stats{'bytes'}+=length($_);
	tr/A-Za-z/ /cs;# alle nicht Woerter loeschen / delete all non-word chars
#	s/[^A-Za-z]+/ /g;# alle nicht Woerter loeschen / delete all non-word chars
	if (${opt{'casesensitive'}}==1) {
		tr/A-Z/a-z/;   # Gross- zu Kleinschreibung wandeln / upper to lower case
		tr///;
	}
#		s//Ae/g; # umlaut substitution because older graphviz could not handle these
#		s//ae/g;
#		s//Oe/g;
#		s//oe/g;
#		s//Ue/g;
#		s//ue/g;
#		s//sz/g;

	#    print "... $_\n";
	foreach $_ (grep {length $_ > 1} split (/ +/)) {
		$stats{'words'}++;
		$stats{'wordchars'}+=length($_);
		if (not defined $words{$_}) {
			$words{$_}=1;
			$stats{'uniqs'}++;
		} else {
			$words{$_}++;
		}
	}
	$stats{'whitespaces'}=$stats{'bytes'} - $stats{'wordchars'};
	if (
		${opt{'with-progress'}} == 1 && 
		$stats{'bytes'} % 4096 == 0 &&
		time - $stats{'starttime'} > 0
	) {
		my $duration=(time - $stats{'starttime'});
		my $d_hours=int($duration / 3600);
		my $d_mins=int(($duration - $d_hours*3600)/60);
		my $d_secs=int(($duration - $d_hours*3600 - $d_mins*60));
		printf(STDERR "\rread Mbytes: %6.2g, MB/s: %.1f, duration: %02d:%02d:%02d, words: %5.2e, uniq: %5.2e, lines: %5.2g, wordratio: %.2f%%",
			$stats{'bytes'}/(1024*1024),
			($stats{'bytes'}/(1024*1024) / (time - $stats{'starttime'})),
			$d_hours,
			$d_mins,
			$d_secs,
			$stats{'words'},
			$stats{'uniqs'},
			$stats{'lines'},
			100*$stats{'wordchars'} / $stats{'bytes'}
		);
	}
}
if (${opt{'wordlen'}}) {
	foreach my $sorted_word (reverse sort {length($a) <=> length($b)} keys %words) {
		printf(OUTFILE "'%s' -> %i (freq: %0.5f)\n", $sorted_word, length($sorted_word), $words{$sorted_word}/$stats{'words'});
	}
} elsif (${opt{'wordfreq'}}) {
	foreach my $sorted_word (reverse sort {$words{$a} <=> $words{$b}} keys %words) {
		printf(OUTFILE "'%s' -> %0.5f\n", $sorted_word, $words{$sorted_word}/$stats{'words'});
	}
}


